"""`AdvancedFilterSet` class module.

Use the `AdvancedFilterSet` class from this module instead of the `FilterSet` from django-filter.
"""

from typing import Any, Dict, List, Optional, Type, Union, cast

from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.forms import Form
from django.forms.utils import ErrorDict
from django_filters import Filter
from django_filters.conf import settings
from django_filters.filterset import BaseFilterSet, FilterSetMetaclass
from graphene.types.inputobjecttype import InputObjectTypeContainer
from wrapt import ObjectProxy


def tree_input_type_to_data(
    tree_input_type: InputObjectTypeContainer,
    prefix: str = '',
) -> Dict[str, Any]:
    """Convert a tree_input_type to a FilterSet data."""
    result: Dict[str, Any] = {}
    for key, value in tree_input_type.items():
        if key in ('and', 'or'):
            result[key] = [tree_input_type_to_data(subtree) for subtree in value]
        elif key == 'not':
            result[key] = tree_input_type_to_data(value)
        else:
            k = (prefix + LOOKUP_SEP + key if prefix else key).replace(
                LOOKUP_SEP + settings.DEFAULT_LOOKUP_EXPR, '',
            )
            if isinstance(value, InputObjectTypeContainer):
                result.update(tree_input_type_to_data(value, k))
            else:
                result[k] = value
    return result


class QuerySetProxy(ObjectProxy):
    """Proxy for a QuerySet object.

    The Django-filter library works with QuerySet objects,
    but such objects do not provide the ability to apply the negation operator to the entire object.
    Therefore, it is convenient to work with the Q object instead of the QuerySet.
    This class replaces the original QuerySet object,
    and creates a Q object when calling the `filter` and `exclude` methods.
    """

    __slots__ = 'q'

    def __init__(self, wrapped: models.QuerySet) -> None:
        super().__init__(wrapped)
        self.q = models.Q()

    def __getattr__(self, name: str) -> Any:
        """Return QuerySet attributes for all cases except `filter` and `exclude`."""
        if name == 'filter':
            return self.filter_
        elif name == 'exclude':
            return self.exclude_
        return super().__getattr__(name)

    def filter_(self, *args, **kwargs) -> 'QuerySetProxy':
        """Replace the `filter` method of the QuerySet class."""
        if len(kwargs) == 0 and len(args) == 1 and isinstance(args[0], models.Q):
            q = args[0]
        else:
            q = models.Q(*args, **kwargs)
        self.q = self.q & q
        return self

    def exclude_(self, *args, **kwargs) -> 'QuerySetProxy':
        """Replace the `exclude` method of the QuerySet class."""
        if len(kwargs) == 0 and len(args) == 1 and isinstance(args[0], models.Q):
            q = args[0]
        else:
            q = models.Q(*args, **kwargs)
        self.q = self.q & ~q
        return self


def get_q(queryset: models.QuerySet, filter_obj: Filter, value: Any) -> models.Q:
    """Return a Q object for a queryset, filter and value."""
    queryset_proxy = QuerySetProxy(queryset)
    filter_obj.filter(queryset_proxy, value)
    return queryset_proxy.q


class AdvancedFilterSet(BaseFilterSet, metaclass=FilterSetMetaclass):
    """Allow you to use advanced filters with `or` and `and` expressions."""

    class TreeFormMixin(Form):
        """Tree-like form mixin."""

        def __init__(
            self,
            and_forms: Optional[List['AdvancedFilterSet.TreeFormMixin']] = None,
            or_forms: Optional[List['AdvancedFilterSet.TreeFormMixin']] = None,
            not_form: Optional['AdvancedFilterSet.TreeFormMixin'] = None,
            *args,
            **kwargs
        ) -> None:
            super().__init__(*args, **kwargs)
            self.and_forms = and_forms or []
            self.or_forms = or_forms or []
            self.not_form = not_form

        @property
        def errors(self) -> ErrorDict:
            """Return an ErrorDict for the data provided for the form."""
            self_errors: ErrorDict = super().errors
            for key in ('and', 'or'):
                errors: ErrorDict = ErrorDict()
                for i, form in enumerate(getattr(self, f'{key}_forms')):
                    if form.errors:
                        errors[f'{key}_{i}'] = form.errors
                if len(errors):
                    self_errors.update({key: errors})
            if self.not_form and self.not_form.errors:
                self_errors.update({'not': self.not_form.errors})
            return self_errors

    def get_form_class(self) -> Type[Union[Form, TreeFormMixin]]:
        """Return a django Form class suitable of validating the filterset data.

        The form must be tree-like because the data is tree-like.
        """
        form_class = super(AdvancedFilterSet, self).get_form_class()
        tree_form = cast(
            Type[Union[Form, AdvancedFilterSet.TreeFormMixin]],
            type(
                f'{form_class.__name__.replace("Form", "")}TreeForm',
                (form_class, AdvancedFilterSet.TreeFormMixin),
                {},
            ),

        )
        return tree_form

    @property
    def form(self) -> Union[Form, TreeFormMixin]:
        """Return a django Form suitable of validating the filterset data."""
        if not hasattr(self, '_form'):
            form_class = self.get_form_class()
            if self.is_bound:
                self._form = self.create_form(form_class, self.data)
            else:
                self._form = form_class(prefix=self.form_prefix)
        return self._form

    def create_form(
        self,
        form_class: Type[Union[Form, TreeFormMixin]],
        data: Dict[str, Any],
    ) -> Union[Form, TreeFormMixin]:
        """Create a form from a form class and data."""
        return form_class(
            data={k: v for k, v in data.items() if k not in ('and', 'or', 'not')},
            and_forms=[self.create_form(form_class, and_data) for and_data in data.get('and', [])],
            or_forms=[self.create_form(form_class, or_data) for or_data in data.get('or', [])],
            not_form=self.create_form(form_class, data['not']) if data.get('not', None) else None,
        )

    def find_filter(self, data_key: str) -> Filter:
        """Find a filter using a data key.

        The data key may differ from a filter name, because
        the data keys may contain DEFAULT_LOOKUP_EXPR and user can create
        a AdvancedFilterSet class without following the naming convention.
        """
        if LOOKUP_SEP in data_key:
            field_name, lookup_expr = data_key.rsplit(LOOKUP_SEP, 1)
        else:
            field_name, lookup_expr = data_key, settings.DEFAULT_LOOKUP_EXPR
        key = field_name if lookup_expr == settings.DEFAULT_LOOKUP_EXPR else data_key
        if key in self.filters:
            return self.filters[key]
        for filter_value in self.filters.values():
            if filter_value.field_name == field_name and filter_value.lookup_expr == lookup_expr:
                return filter_value

    def filter_queryset(self, queryset: models.QuerySet) -> models.QuerySet:
        """Filter a queryset with a top level form's `cleaned_data`."""
        return queryset.filter(self.get_q_for_form(queryset, self.form))

    def get_q_for_form(
        self,
        queryset: models.QuerySet,
        form: Union[Form, TreeFormMixin],
    ) -> models.Q:
        """Return a Q object for a form's `cleaned_data` using `and`, `or` or `not` operator."""
        q = models.Q()
        for name, value in form.cleaned_data.items():
            q = q & get_q(queryset, self.find_filter(name), value)
        and_q = models.Q()
        for and_form in form.and_forms:
            and_q = and_q & self.get_q_for_form(queryset, and_form)
        or_q = models.Q()
        for or_form in form.or_forms:
            or_q = or_q | self.get_q_for_form(queryset, or_form)
        not_q = ~self.get_q_for_form(queryset, form.not_form) if form.not_form else models.Q()
        return q & and_q & or_q & not_q



"""`AdvancedDjangoFilterConnectionField` class module.

Use the `AdvancedDjangoFilterConnectionField` class from this
module instead of the `DjangoFilterConnectionField` from graphene-django.
"""

from typing import Any, Dict, Iterable, Optional, Type

import graphene
from django.core.exceptions import ValidationError
from django.db import models
from django_filters import FilterSet
from graphene_django import DjangoObjectType
from graphene_django.filter import DjangoFilterConnectionField

from .filterset import AdvancedFilterSet, tree_input_type_to_data
from .filterset_factories import get_filterset_class
from .input_type_factories import get_filtering_args_from_filterset


class AdvancedDjangoFilterConnectionField(DjangoFilterConnectionField):
    """Allow you to use advanced filters provided by this library."""

    def __init__(
        self,
        type: Type[DjangoObjectType],
        fields: Optional[Dict[str, list]] = None,
        order_by: Any = None,
        extra_filter_meta: Optional[dict] = None,
        filterset_class: Optional[Type[AdvancedFilterSet]] = None,
        *args,
        **kwargs
    ) -> None:
        assert filterset_class is None or issubclass(filterset_class, AdvancedFilterSet), \
            'Use the `AdvancedFilterSet` class with the `AdvancedDjangoFilterConnectionField`'
        super().__init__(
            type,
            fields,
            order_by,
            extra_filter_meta,
            filterset_class,
            *args,
            **kwargs
        )

    @property
    def filterset_class(self) -> Type[AdvancedFilterSet]:
        """Return a AdvancedFilterSet instead of a FilterSet."""
        if not self._filterset_class:
            fields = self._fields or self.node_type._meta.filter_fields
            meta = {'model': self.model, 'fields': fields}
            if self._extra_filter_meta:
                meta.update(self._extra_filter_meta)
            filterset_class = self._provided_filterset_class or (
                self.node_type._meta.filterset_class
            )
            self._filterset_class = get_filterset_class(filterset_class, **meta)
        return self._filterset_class

    @property
    def filtering_args(self) -> dict:
        """Return filtering args from the filterset."""
        if not self._filtering_args:
            self._filtering_args = get_filtering_args_from_filterset(
                self.filterset_class, self.node_type,
            )
        return self._filtering_args

    @classmethod
    def resolve_queryset(
        cls,
        connection: object,
        iterable: Iterable,
        info: graphene.ResolveInfo,
        args: Dict[str, Any],
        filtering_args: Dict[str, graphene.InputField],
        filterset_class: Type[FilterSet],
    ) -> models.QuerySet:
        """Return a filtered QuerySet."""
        qs = super(DjangoFilterConnectionField, cls).resolve_queryset(
            connection, iterable, info, args,
        )
        filterset = filterset_class(
            data=tree_input_type_to_data(args['filter']), queryset=qs, request=info.context,
        )
        if filterset.form.is_valid():
            return filterset.qs
        raise ValidationError(filterset.form.errors.as_json())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graphene_django_filter']

package_data = \
{'': ['*']}

install_requires = \
['Django==3.2',
 'anytree>=2.8.0,<3.0.0',
 'django-filter>=21.1,<22.0',
 'django-seed>=0.3.1,<0.4.0',
 'graphene-django>=2.15.0,<3.0.0',
 'graphene==2.1.9',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'stringcase>=1.2.0,<2.0.0',
 'wrapt>=1.13.3,<2.0.0']

setup_kwargs = {
    'name': 'graphene-django-filter',
    'version': '0.5.0',
    'description': 'Advanced filters for Graphene',
    'long_description': '# Graphene-Django-Filter\n[![CI](https://github.com/devind-team/graphene-django-filter/workflows/CI/badge.svg)](https://github.com/devind-team/graphene-django-filter/actions) [![PyPI version](https://badge.fury.io/py/graphene-django-filter.svg)](https://badge.fury.io/py/graphene-django-filter)\n\nThis package contains advanced filters for [graphene-django](https://github.com/graphql-python/graphene-django). The standard filtering feature in graphene-django relies on the [django-filter](https://github.com/carltongibson/django-filter) library and therefore provides the flat API without the ability to use logical operators such as `and`, `or` and `not`. This library makes the API nested and adds logical expressions by extension of the `DjangoFilterConnectionField` field and the `FilterSet` class.\n# Requirements\n* Python (3.6, 3.7, 3.8, 3.9, 3.10)\n* Graphene-Django (2.15)\n# Features\n## Nested API with the ability to use logical operators\nTo use, simply replace all `DjangoFilterConnectionField` fields with `AdvancedDjangoFilterConnectionField` fields in your queries. Also, if you create custom FilterSets, replace the inheritance from the `FilterSet` class with the inheritance from the `AdvancedFilterSet` class. For example, the following task query exposes the old flat API.\n```python\nimport graphene\nfrom django_filters import FilterSet\nfrom graphene_django import DjangoObjectType\nfrom graphene_django.filter import DjangoFilterConnectionField\n\nclass TaskFilter(FilterSet)\n    class Meta:\n        model = Task\n        fields = {\n            \'name\': (\'exact\', \'contains\'),\n            \'user__email\': (\'exact\', \'contains\'),\n            \'user__first_name\': (\'exact\', \'contains\'),\n            \'user__last_name\': (\'exact\', \'contains\'),\n        }\n \nclass UserType(DjangoObjectType):\n    class Meta:\n        model = User\n        interfaces = (graphene.relay.Node,)\n        fields = \'__all__\'\n        \nclass TaskType(DjangoObjectType):\n    user = graphene.Field(UserType)\n\n    class Meta:\n        model = Task\n        interfaces = (graphene.relay.Node,)\n        fields = \'__all__\'\n        filterset_class = TaskFilter\n        \nclass Query(graphene.ObjectType):\n    tasks = DjangoFilterConnectionField(TaskType)\n```\nThe flat API in which all filters are applied using the `and` operator looks like this.\n```graphql\n{\n  tasks(\n    name_Contains: "important"\n    user_Email_Contains: "john"\n    user_FirstName: "John"\n    user_LastName: "Dou"\n  ){\n    edges {\n      node {\n        id\n        name\n      }\n    }\n  }\n}\n```\nAfter replacing the field class with the `AdvancedDjangoFilterConnectionField` and the `FilterSet` class with the `AdvancedFilterSet` the API becomes nested with support for logical expressions.\n```python\nfrom graphene_django_filter import AdvancedDjangoFilterConnectionField, AdvancedFilterSet\n\nclass TaskFilter(AdvancedFilterSet)\n    class Meta:\n        model = Task\n        fields = {\n            \'name\': (\'exact\', \'contains\'),\n            \'user__email\': (\'exact\', \'contains\'),\n            \'user__first_name\': (\'exact\', \'contains\'),\n            \'user__last_name\': (\'exact\', \'contains\'),\n        }\n\nclass Query(graphene.ObjectType):\n    tasks = AdvancedDjangoFilterConnectionField(TaskType)\n```\nFor example, the following query returns tasks which names contain the word "important" or the user\'s email address contains the word "john" and the user\'s last name is "Dou" and the first name is not "John". Note that the operators are applied to lookups such as `contains`, `exact`, etc. at the last level of nesting.\n```graphql\n{\n  tasks(\n    filter: {\n      or: [\n        {name: {contains: "important"}}\n        and: [\n          {user: {email: {contains: "john"}}}\n          {user: {lastName: {exact: "Dou"}}}\n        ]\n      ]\n      not: {\n        {user: {firstName: {exact: "John"}}}\n      }\n    }\n  ){\n    edges {\n      node {\n        id\n        name\n      }\n    }\n  }\n}\n```\nThe same result can be achieved with an alternative query structure because within the same object the `and` operator is always used.\n```graphql\n{\n  tasks(\n    filter: {\n      or: [\n        {name: {contains: "important"}}\n        {\n          user: {\n            email: {contains: "john"}\n            lastName: {exact: "Dou"}\n          }\n        }\n      ]\n      not: {\n        {user: {firstName: {exact: "John"}}}\n      }\n    }\n  ){\n    edges {\n      node {\n        id\n        name\n      }\n    }\n  }\n}\n```\nThe filter input type has the following structure.\n```graphql\ninput FilterInputType {\n  and: [FilterInputType]\n  or: [FilterInputType]\n  not: FilterInputType\n  ...FieldLookups\n}\n```\nFor more examples, see [tests](https://github.com/devind-team/graphene-django-filter/blob/8faa52bdfc2a66fc74a8aecb798b8358f7f7ea7c/tests/test_connection_field.py#L19).\n',
    'author': 'devind-team',
    'author_email': 'team@devind.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/devind-team/graphene-django-filter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2',
}


setup(**setup_kwargs)



"""Functions for converting a FilterSet class to a tree and then to an input type."""

from typing import Any, Dict, List, Optional, Sequence, Type, cast

import graphene
from anytree import Node
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django_filters import Filter
from django_filters.conf import settings
from graphene_django.filter.utils import get_model_field
from graphene_django.forms.converter import convert_form_field
from stringcase import pascalcase

from .filterset import AdvancedFilterSet


def get_filtering_args_from_filterset(
    filterset_class: Type[AdvancedFilterSet],
    node_type: Type[graphene.ObjectType],
) -> Dict[str, graphene.Argument]:
    """Inspect a FilterSet and produce the arguments to pass to a Graphene Field.

    These arguments will be available to filter against in the GraphQL.
    """
    return {
        'filter': graphene.Argument(
            create_filter_input_type(
                filterset_to_trees(filterset_class),
                filterset_class,
                node_type.__name__.replace('Type', ''),
            ), description='Advanced filter field',
        ),
    }


def create_filter_input_type(
    roots: List[Node],
    filterset_class: Type[AdvancedFilterSet],
    type_name: str,
) -> Type[graphene.InputObjectType]:
    """Create a filter input type from filter set trees."""
    input_type = cast(
        Type[graphene.InputObjectType],
        type(
            f'{type_name}FilterInputType',
            (graphene.InputObjectType,),
            {
                **{
                    root.name: graphene.InputField(
                        create_filter_input_subtype(root, filterset_class, type_name),
                        description=f'`{pascalcase(root.name)}` field',
                    )
                    for root in roots
                },
                'and': graphene.InputField(
                    graphene.List(lambda: input_type),
                    description='`And` field',
                ),
                'or': graphene.InputField(
                    graphene.List(lambda: input_type),
                    description='`Or` field',
                ),
                'not': graphene.InputField(lambda: input_type, description='`Not` field'),
            },
        ),
    )
    return input_type


def create_filter_input_subtype(
    root: Node,
    filterset_class: Type[AdvancedFilterSet],
    prefix: str,
) -> Type[graphene.InputObjectType]:
    """Create a filter input subtype from a filter set subtree."""
    fields: Dict[str, graphene.InputField] = {}
    for child in root.children:
        if child.height == 0:
            filter_name = f'{LOOKUP_SEP}'.join(
                node.name for node in child.path if node.name != settings.DEFAULT_LOOKUP_EXPR
            )
            fields[child.name] = get_field(
                filterset_class,
                filter_name,
                filterset_class.base_filters[filter_name],
            )
        else:
            fields[child.name] = graphene.InputField(
                create_filter_input_subtype(
                    child,
                    filterset_class,
                    prefix + pascalcase(root.name),
                ),
                description=f'`{pascalcase(child.name)}` subfield',
            )
    return create_input_object_type(
        f'{prefix}{pascalcase(root.name)}FilterInputType',
        fields,
    )


def create_input_object_type(name: str, fields: Dict[str, Any]) -> Type[graphene.InputObjectType]:
    """Create an inheritor for the `InputObjectType` class."""
    return cast(
        Type[graphene.InputObjectType],
        type(
            name,
            (graphene.InputObjectType,),
            fields,
        ),
    )


def get_field(
    filterset_class: Type[AdvancedFilterSet],
    name: str,
    filter_field: Filter,
) -> graphene.InputField:
    """Return Graphene input field from a filter field.

    It is a partial copy of the `get_filtering_args_from_filterset` function from graphene-django.
    https://github.com/graphql-python/graphene-django/blob/caf954861025b9f3d9d3f9c204a7cbbc87352265/graphene_django/filter/utils.py#L11
    """
    model = filterset_class._meta.model
    form_field: Optional[models.Field] = None
    filter_type: str = filter_field.lookup_expr
    if name in getattr(filterset_class, 'declared_filters'):
        form_field = filter_field.field
        field = convert_form_field(form_field)
    else:
        model_field = get_model_field(model, filter_field.field_name)
        if filter_type != 'isnull' and hasattr(model_field, 'formfield'):
            form_field = model_field.formfield(
                required=filter_field.extra.get('required', False),
            )
        if not form_field:
            form_field = filter_field.field
        field = convert_form_field(form_field)
    if filter_type in ('in', 'range'):
        field = graphene.List(field.get_type())
    field_type = field.InputField()
    field_type.description = getattr(filter_field, 'label') or \
        f'`{pascalcase(filter_field.lookup_expr)}` lookup'
    return field_type


def filterset_to_trees(filterset_class: Type[AdvancedFilterSet]) -> List[Node]:
    """Convert a FilterSet class to trees."""
    trees: List[Node] = []
    for filter_value in filterset_class.base_filters.values():
        values = (*filter_value.field_name.split(LOOKUP_SEP), filter_value.lookup_expr)
        if len(trees) == 0 or not any([try_add_sequence(tree, values) for tree in trees]):
            trees.append(sequence_to_tree(values))
    return trees


def try_add_sequence(root: Node, values: Sequence[str]) -> bool:
    """Try to add a sequence to a tree.

    Return a flag indicating whether the mutation was made.
    """
    if root.name == values[0]:
        for child in root.children:
            is_mutated = try_add_sequence(child, values[1:])
            if is_mutated:
                return True
        root.children = (*root.children, sequence_to_tree(values[1:]))
        return True
    else:
        return False


def sequence_to_tree(values: Sequence[str]) -> Node:
    """Convert a sequence to a tree."""
    node: Optional[Node] = None
    for value in values:
        node = Node(name=value, parent=node)
    return node.root
